import os
import urllib.error
import urllib.parse
import urllib.parse
import urllib.request

from OpenSSL import crypto
from suds.transport import Reply, Transport
from suds.transport.sslverify import optionsForClientTLS
from twisted.internet import defer, reactor
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.protocol import Protocol
from twisted.web import http
from twisted.web.client import Agent, ProxyAgent, _requireSSL
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer, IPolicyForHTTPS
from zope.interface import implementer


class BytesResponseConsumer(Protocol):
    """
    Protocol that consumes the entire response body into a string and provides
    a simple callback interface for the user to be triggered when the response
    is complete.

    @ivar response:  The response that filled us.
    @ivar _finished: Deferred that is triggered when the body is completed.
    """

    def __init__(self):
        self._finished = defer.Deferred()
        self.response = None
        self.body = b""

    def getDeferred(self):
        """ Return the deferred that is triggered after full completion. """
        return self._finished

    def dataReceived(self, data):
        self.body += data

    def connectionLost(self, reason):
        """ Callback to finished with copy of ourselves. """
        self._finished.callback(self)

    def responseWithoutBody(self):
        """ Called when the response does not contain a body. """
        self._finished.callback(self)


@implementer(IBodyProducer)
class BytesProducer(object):
    """
    Simple wrapper around a string that will produce that string with the correct
    interface.
    """

    def __init__(self, body: bytes):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)

        return defer.succeed(None)

    def pauseProducing(self):
        pass

    def resumeProducing(self):
        pass

    def stopProducing(self):
        pass


@implementer(IPolicyForHTTPS)
class PolicyForHTTPS(object):
    """
    Custom SSL connection creator that allows specifying private key, certificate
    and custom options.
    """

    def __init__(self, trustRoot=None, privateKey=None, certificate=None, **kwargs):
        self._trustRoot = trustRoot
        self._opts = dict(kwargs)
        self._opts["privateKey"] = privateKey
        self._opts["certificate"] = certificate

    @_requireSSL
    def creatorForNetloc(self, hostname, port):
        """
        Create a client connection creator for a given network location.

        @param hostname: The hostname part of the URI.
        @type  hostname: L{bytes}
        @param port:     The port part of the URI.
        @type  port:     L{int}
        """
        return optionsForClientTLS(hostname.decode("ascii"),
                                   trustRoot=self._trustRoot,
                                   extraCertificateOptions=self._opts)


class TwistedTransport(Transport):
    """
    Custom transport that uses the Twisted REST client.
    """

    def __init__(self):
        """
        Constructor.
        """
        Transport.__init__(self)
        from suds.transport.options import Options
        self.options = Options()
        del Options
        self._httpsPolicy = None

    def _getHttpsPolicy(self):
        """
        Helper method that lazily constructs the HTTPS options to use for this
        transport.
        """
        if self._httpsPolicy is not None:
            return self._httpsPolicy

        # Attempt to load the certificate and private key from a file.
        certificate = None
        if self.options.certificate:
            cert_data = self.options.certificate
            if os.path.isfile(cert_data):
                with open(cert_data, "rb") as cert_file:
                    cert_data = cert_file.read()
            certificate = crypto.load_certificate(crypto.FILETYPE_PEM, cert_data)
        priv_key = None
        if self.options.privateKey:
            key_data = self.options.privateKey
            if os.path.isfile(key_data):
                with open(key_data, "rb") as key_file:
                    key_data = key_file.read()
            priv_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_data)

        # Get the rest of the options for the context factory.
        other_opts = {}
        for opt_name in ['method', 'verify', 'caCerts', 'verifyDepth', 'trustRoot',
                         'requireCertificate', 'verifyOnce', 'enableSingleUseKeys',
                         'enableSessions', 'fixBrokenPeers', 'enableSessionTickets',
                         'acceptableCiphers']:
            other_opts[opt_name] = getattr(self.options, opt_name)

        self._httpsPolicy = PolicyForHTTPS(privateKey=priv_key,
                                           certificate=certificate,
                                           **other_opts)
        return self._httpsPolicy

    httpsPolicy = property(_getHttpsPolicy)

    @defer.inlineCallbacks
    def _request(self, request, method):
        """
        Helper method that sends the given HTTP request.
        """
        # Copy the headers from the request.
        headers = Headers()
        for (key, value) in request.headers.items():
            headers.addRawHeader(key, value)

        # If a username and password are given, then add basic authentication.
        if (self.options.username is not None and
                    self.options.password is not None):
            auth = "%s:%s" % (self.options.username, self.options.password)
            auth = auth.encode("base64").strip()
            headers.addRawHeader(b'Authorization', b'Basic ' + auth)

        # Determine if the user has configured a proxy server.
        url_parts = urllib.parse.urlparse(request.url)
        proxy = self.options.proxy.get(url_parts.scheme, None)

        # Construct an agent to send the request.
        if proxy is not None:
            (hostname, port) = proxy.split(":")
            endpoint = TCP4ClientEndpoint(reactor, hostname, int(port),
                                          timeout=self.options.timeout)
            agent = ProxyAgent(endpoint)
        else:
            agent = Agent(reactor, self.httpsPolicy,
                          connectTimeout=self.options.timeout)

        url = request.url.encode("utf-8")
        producer = BytesProducer(request.message or b"")
        response = yield agent.request(method.encode(),
                                       url, headers, producer)

        # If the initial request returned a redirection response, attempt to follow it.
        http_redirect_codes = [http.MOVED_PERMANENTLY,  # 301
                               http.FOUND,  # 302
                               http.SEE_OTHER,  # 303
                               http.TEMPORARY_REDIRECT]  # 307
        if response.code in http_redirect_codes and response.headers.hasHeader(
                "Location"):
            new_url = response.headers.getRawHeaders("Location")[0]
            producer = BytesProducer(request.message or "")
            response = yield agent.request(method, new_url, headers, producer)

        # Construct a simple response consumer and give it the response body.
        consumer = BytesResponseConsumer()
        response.deliverBody(consumer)
        yield consumer.getDeferred()
        consumer.response = response
        defer.returnValue(consumer)

    @defer.inlineCallbacks
    def open(self, request):
        """
        Open the url in the specified request.

        @param request: A transport request.
        @type  request: L{Request}

        @return: An input stream.
        @rtype:  stream

        @raise TransportError: On all transport errors.
        """
        if request.url.startswith("file://"):
            url_parts = urllib.parse.urlparse(request.url)
            full_path = os.path.join(url_parts.netloc, url_parts.path)
            local_fname = urllib.request.url2pathname(full_path)

            with open(local_fname, "rb") as local_file:
                content = local_file.read()
            defer.returnValue(content)

        consumer = yield self._request(request, "GET")
        defer.returnValue(consumer.body)

    @defer.inlineCallbacks
    def send(self, request):
        """
        Send soap message.  Implementations are expected to handle:
            - proxies
            - I{http} headers
            - cookies
            - sending message
            - brokering exceptions into L{TransportError}

        @param request: A transport request.
        @type request: L{Request}
        @return: The reply
        @rtype: L{Reply}
        @raise TransportError: On all transport errors.
        """
        consumer = yield self._request(request, "POST")
        res_headers = {}
        for key, values in consumer.response.headers.getAllRawHeaders():
            res_headers[key.decode()] = [v.decode() for v in values]
        result = Reply(consumer.response.code,
                       res_headers,
                       consumer.body.decode())
        defer.returnValue(result)



import logging
import re

from OpenSSL import SSL
from OpenSSL._util import (
    lib as _lib,
)

SSL_CB_HANDSHAKE_START = _lib.SSL_CB_HANDSHAKE_START
SSL_CB_HANDSHAKE_DONE = _lib.SSL_CB_HANDSHAKE_DONE

from twisted.python.failure import Failure

from twisted.internet.ssl import Certificate
from twisted.internet._sslverify import (ClientTLSOptions, OpenSSLCertificateOptions)


def _maybeSetHostNameIndication(connection, hostname):
    if getattr(SSL.Connection, "set_tlsext_host_name", None) is not None:
        connection.set_tlsext_host_name(hostname)


log = logging.getLogger(__name__)


def optionsForClientTLS(hostname, trustRoot=None, clientCertificate=None, **kw):
    """
    Reimplemented from twisted.internet.ssl to allow extra parameters to be passed correctly.

    @return: A client connection creator.
    @rtype: L{IOpenSSLClientConnectionCreator}
    """
    extraCertificateOptions = kw.pop('extraCertificateOptions', None) or {}
    if kw:
        raise TypeError(
            "optionsForClientTLS() got an unexpected keyword argument"
            " '{arg}'".format(
                arg=kw.popitem()[0]
            )
        )
    if not isinstance(hostname, str):
        raise TypeError(
            "optionsForClientTLS requires text for host names, not "
            + hostname.__class__.__name__
        )
    if clientCertificate:
        extraCertificateOptions.update(
            privateKey=clientCertificate.privateKey.original,
            certificate=clientCertificate.original
        )

    # Only pass the trustRoot if it is not None to avoid mutually exclusive param issues.
    if trustRoot:
        certificateOptions = OpenSSLCertificateOptions(
            trustRoot=trustRoot,
            **extraCertificateOptions
        )
    else:
        certificateOptions = OpenSSLCertificateOptions(**extraCertificateOptions)

    return SSLClientConnectionCreator(hostname, certificateOptions.getContext())


class CertMatchError(ValueError):
    """
    Raised when we fail to match an SSL certificate to a hostname.
    """
    pass


def _dnsNameMatch(name, hostname, maxWildcards=1):
    """
    Matching according to RFC 6125, section 6.4.3
    @type  name:         basestring
    @param name:         The name/value to check against the hostname.
    @type  hostname:     basestring
    @param hostname:     Hostname to check against.
    @type  maxWildcards: int
    @param maxWildcards: The maximum number of wildcards to allow in name.
    http://tools.ietf.org/html/rfc6125#section-6.4.3
    """
    pats = []
    if not name:
        return False

    # Ported from python3-syntax:
    # leftmost, *remainder = name.split(r'.')
    parts = name.split(r".")
    leftmost = parts[0]
    remainder = parts[1:]

    wildcards = leftmost.count("*")
    if wildcards > maxWildcards:
        # Issue #17980: avoid denials of service by refusing more
        # than one wildcard per fragment.  A survey of established
        # policy among SSL implementations showed it to be a
        # reasonable choice.
        raise CertMatchError("Too many wildcards in certificate DNS name: %s" % (name))

    # speed up common case w/o wildcards
    if not wildcards:
        return name.lower() == hostname.lower()

    # RFC 6125, section 6.4.3, subitem 1.
    # The client SHOULD NOT attempt to match a presented identifier in which
    # the wildcard character comprises a label other than the left-most label.
    if leftmost == "*":
        # When "*" is a fragment by itself, it matches a non-empty dotless
        # fragment.
        pats.append("[^.]+")
    elif leftmost.startswith("xn--") or hostname.startswith("xn--"):
        # RFC 6125, section 6.4.3, subitem 3.
        # The client SHOULD NOT attempt to match a presented identifier
        # where the wildcard character is embedded within an A-label or
        # U-label of an internationalized domain name.
        pats.append(re.escape(leftmost))
    else:
        # Otherwise, "*" matches any dotless string, e.g. www*
        pats.append(re.escape(leftmost).replace(r"\*", "[^.]*"))

    # add the remaining fragments, ignore any wildcards
    for frag in remainder:
        pats.append(re.escape(frag))

    return re.match(r"\A" + r"\.".join(pats) + r"\Z", hostname, re.IGNORECASE) is not None


def _matchHostname(cert, hostname):
    """
    The match_hostname() function from Python 3.3.3
    Verify that L{cert} matches the L{hostname}. RFC 2818 and RFC 6125 rules
    are followed, but IP addresses are not accepted for L{hostname}.
    @type  cert:     L{twisted.internet.ssl.Certificate}
    @param cert:     SSL certificate to verify hostname for.
    @type  hostname: basestring
    @param hostname: Hostname to check against.
    CertMatchError is raised on failure. On success, the function returns nothing.
    """
    # The common name is only checked when there are no DNS entries in the SAN.
    common_name = cert.original.get_subject().commonName
    if not _dnsNameMatch(common_name, hostname):
        raise CertMatchError("Hostname %s doesn't match %s" % (hostname, common_name))


class SSLClientConnectionCreator(ClientTLSOptions):
    """
    Client creator for TLS.

    Extends ClientTLSOptions to allow skipping verification and to improve
    twisted's base verification.
    """

    def _identityVerifyingInfoCallback(self, connection, where, ret):
        """
        Override the base implementation to provide better hostname verification.

        @param connection: the connection which is handshaking.
        @type connection: L{OpenSSL.SSL.Connection}

        @param where: flags indicating progress through a TLS handshake.
        @type where: L{int}

        @param ret: ignored
        @type ret:  ignored
        """
        if where & SSL_CB_HANDSHAKE_START:
            _maybeSetHostNameIndication(connection, self._hostnameBytes)
        elif where & SSL_CB_HANDSHAKE_DONE:
            if self._ctx.get_verify_mode() != SSL.VERIFY_NONE:
                try:
                    peer_cert = Certificate(connection.get_peer_certificate())
                    _matchHostname(peer_cert, self._hostname)
                except CertMatchError as ex:
                    log.error(str(ex))
                    f = Failure()
                    transport = connection.get_app_data()
                    transport.failVerification(f)
